@@ -156,7 +156,7 @@ class IDatabaseIterator(Interface):
 class ILocker(Interface):
     """Acquire and release the commit and pack locks."""
 
-    def hold_commit_lock(cursor, ensure_current=False):
+    def hold_commit_lock(cursor, ensure_current=False, nowait=False):
         """Acquire the commit lock.
 
         If ensure_current is True, other tables may be locked as well, to
@@ -164,6 +164,10 @@ class ILocker(Interface):
 
         May raise StorageError if the lock can not be acquired before
         some timeout.
+
+        With nowait set to True, only try to obtain the lock without waiting
+        and return a boolean indicating if the lock was successful.
+
         """
 
     def release_commit_lock(cursor):
@@ -253,7 +257,7 @@ class IObjectMover(Interface):
         This happens after conflict resolution.
         """
 
-    def move_from_temp(cursor, tid):
+    def move_from_temp(cursor, tid, txn_has_blobs):
         """Moved the temporarily stored objects to permanent storage.
 
         Returns the list of oids stored.
@@ -265,6 +269,18 @@ class IObjectMover(Interface):
         tid is the integer tid of the transaction being committed.
         """
 
+    def download_blob(cursor, oid, tid, filename):
+        """Download a blob into a file.
+
+        Returns the size of the blob file in bytes.
+        """
+
+    def upload_blob(cursor, oid, tid, filename):
+        """Upload a blob from a file.
+
+        If tid is None, upload to the temporary table.
+        """
+
 
 class IOIDAllocator(Interface):
     """Allocate OIDs and control future allocation"""
@@ -305,20 +321,16 @@ class IPackUndo(Interface):
         Returns None if there is nothing to pack.
         """
 
-    def pre_pack(pack_tid, get_references, options):
+    def pre_pack(pack_tid, get_references):
         """Decide what to pack.
 
         pack_tid specifies the most recent transaction to pack.
 
         get_references is a function that accepts a stored object state
         and returns a set of OIDs that state refers to.
-
-        options is an instance of relstorage.Options.
-        In particular, the options.pack_gc flag indicates whether
-        to run garbage collection.
         """
 
-    def pack(pack_tid, options, sleep=None, packed_func=None):
+    def pack(pack_tid, sleep=None, packed_func=None):
         """Pack.  Requires the information provided by pre_pack.
 
         packed_func, if provided, will be called for every object state
@@ -326,8 +338,8 @@ class IPackUndo(Interface):
         accept two parameters, oid and tid (64 bit integers).
 
         The sleep function defaults to time.sleep(). It can be
-        overridden to do something else instead of sleep during pauses
-        configured by the duty cycle.
+        overridden to do something else instead of sleep during
+        pauses.
         """
 
 
