@@ -14,6 +14,9 @@
 """Pack/Undo implementations.
 """
 
+from base64 import decodestring
+from itertools import groupby
+from operator import itemgetter
 from relstorage.adapters.interfaces import IPackUndo
 from ZODB.POSException import UndoError
 from zope.interface import implements
@@ -28,10 +31,12 @@ class PackUndo(object):
 
     verify_sane_database = False
 
-    def __init__(self, connmanager, runner, locker):
+    def __init__(self, database_name, connmanager, runner, locker, options):
+        self.database_name = database_name
         self.connmanager = connmanager
         self.runner = runner
         self.locker = locker
+        self.options = options
 
     def choose_pack_transaction(self, pack_point):
         """Return the transaction before or at the specified pack time.
@@ -50,81 +55,97 @@ class PackUndo(object):
         finally:
             self.connmanager.close(conn, cursor)
 
-    def _visit_all_references(self, cursor):
-        """Visit all references in pack_object and set the keep flags.
+    def _traverse_graph(self, cursor):
+        """Visit the entire object graph to find out what should be kept.
+
+        Sets the pack_object.keep flags.
         """
-        # Each of the objects to be kept might refer to other objects.
-        # Mark the referenced objects to be kept as well. Do this
-        # repeatedly until all references have been satisfied.
-        pass_num = 1
-        while True:
-            log.info("pre_pack: following references, pass %d", pass_num)
+        log.info("pre_pack: downloading pack_object and object_ref.")
 
-            # Make a list of all parent objects that still need to be
-            # visited. Then set pack_object.visited for all pack_object
-            # rows with keep = true.
+        # Download the list of root objects to keep from pack_object.
+        keep_set = set()  # set([oid])
+        stmt = """
+        SELECT zoid
+        FROM pack_object
+        WHERE keep = %(TRUE)s
+        """
+        self.runner.run_script_stmt(cursor, stmt)
+        for from_oid, in cursor:
+            keep_set.add(from_oid)
+
+        # Download the list of object references into all_refs.
+        all_refs = {}  # {from_oid: set([to_oid])}
+        # Note the Oracle optimizer hints in the following statement; MySQL
+        # and PostgreSQL ignore these. Oracle fails to notice that pack_object
+        # is now filled and chooses the wrong execution plan, completely
+        # killing this query on large RelStorage databases, unless these hints
+        # are included.
+        stmt = """
+        SELECT 
+            /*+ FULL(object_ref) */ 
+            /*+ FULL(pack_object) */ 
+            object_ref.zoid, object_ref.to_zoid
+        FROM object_ref
+            JOIN pack_object ON (object_ref.zoid = pack_object.zoid)
+        WHERE object_ref.tid >= pack_object.keep_tid
+        ORDER BY object_ref.zoid
+        """
+        self.runner.run_script_stmt(cursor, stmt)
+        # Grouped by object_ref.zoid, store all object_ref.to_zoid in sets
+        for from_oid, rows in groupby(cursor, itemgetter(0)):
+            all_refs[from_oid] = set(row[1] for row in rows)
+
+        # Traverse the object graph.  Add all of the reachable OIDs
+        # to keep_set.
+        log.info("pre_pack: traversing the object graph "
+            "to find reachable objects.")
+        parents = set()
+        parents.update(keep_set)
+        pass_num = 0
+        while parents:
+            pass_num += 1
+            children = set()
+            for parent in parents:
+                to_oids = all_refs.get(parent)
+                if to_oids:
+                    children.update(to_oids)
+            parents = children.difference(keep_set)
+            keep_set.update(parents)
+            log.info("pre_pack: found %d more referenced object(s) in "
+                "pass %d", len(parents), pass_num)
+
+        # Set pack_object.keep for all OIDs in keep_set.
+        del all_refs  # Free some RAM
+        keep_list = list(keep_set)
+        keep_list.sort()
+        log.info("pre_pack: marking objects reachable: %d", len(keep_list))
+        batch = []
+
+        def upload_batch():
+            oids_str = ','.join(str(oid) for oid in batch)
+            del batch[:]
             stmt = """
-            %(TRUNCATE)s temp_pack_visit;
+            UPDATE pack_object SET keep = %%(TRUE)s, visited = %%(TRUE)s
+            WHERE zoid IN (%s)
+            """ % oids_str
+            self.runner.run_script_stmt(cursor, stmt)
 
-            INSERT INTO temp_pack_visit (zoid, keep_tid)
-            SELECT zoid, keep_tid
-            FROM pack_object
-            WHERE keep = %(TRUE)s
-                AND visited = %(FALSE)s;
+        for oid in keep_list:
+            batch.append(oid)
+            if len(batch) >= 1000:
+                upload_batch()
+        if batch:
+            upload_batch()
 
-            UPDATE pack_object SET visited = %(TRUE)s
-            WHERE keep = %(TRUE)s
-                AND visited = %(FALSE)s
-            """
-            self.runner.run_script(cursor, stmt)
-            visit_count = cursor.rowcount
-
-            if self.verify_sane_database:
-                # Verify the update actually worked.
-                # MySQL 5.1.23 fails this test; 5.1.24 passes.
-                stmt = """
-                SELECT 1
-                FROM pack_object
-                WHERE keep = %(TRUE)s AND visited = %(FALSE)s
-                """
-                self.runner.run_script_stmt(cursor, stmt)
-                if list(cursor):
-                    raise AssertionError(
-                        "database failed to update pack_object")
-
-            log.debug("pre_pack: checking references from %d object(s)",
-                visit_count)
-
-            # Visit the children of all parent objects that were
-            # just visited.
-            stmt = self._script_pre_pack_follow_child_refs
-            self.runner.run_script(cursor, stmt)
-            found_count = cursor.rowcount
-
-            log.debug("pre_pack: found %d more referenced object(s) in "
-                "pass %d", found_count, pass_num)
-            if not found_count:
-                # No new references detected.
-                break
-            else:
-                pass_num += 1
-
-    def _pause_pack(self, sleep, options, start):
-        """Pause packing to allow concurrent commits."""
+    def _pause_pack_until_lock(self, cursor, sleep):
+        """Pause until we can obtain a nowait commit lock."""
         if sleep is None:
             sleep = time.sleep
-        elapsed = time.time() - start
-        if elapsed == 0.0:
-            # Compensate for low timer resolution by
-            # assuming that at least 10 ms elapsed.
-            elapsed = 0.01
-        duty_cycle = options.pack_duty_cycle
-        if duty_cycle > 0.0 and duty_cycle < 1.0:
-            delay = min(options.pack_max_delay,
-                elapsed * (1.0 / duty_cycle - 1.0))
-            if delay > 0:
-                log.debug('pack: sleeping %.4g second(s)', delay)
-                sleep(delay)
+        delay = self.options.pack_commit_busy_delay
+        while not self.locker.hold_commit_lock(cursor, nowait=True):
+            cursor.connection.rollback()
+            log.debug('pack: commit lock busy, sleeping %.4g second(s)', delay)
+            sleep(delay)
 
 
 class HistoryPreservingPackUndo(PackUndo):
@@ -145,20 +166,10 @@ class HistoryPreservingPackUndo(PackUndo):
     _script_create_temp_pack_visit = """
         CREATE TEMPORARY TABLE temp_pack_visit (
             zoid BIGINT NOT NULL,
-            keep_tid BIGINT
+            keep_tid BIGINT NOT NULL
         );
-        CREATE UNIQUE INDEX temp_pack_visit_zoid ON temp_pack_visit (zoid)
-        """
-
-    _script_pre_pack_follow_child_refs = """
-        UPDATE pack_object SET keep = %(TRUE)s
-        WHERE keep = %(FALSE)s
-            AND zoid IN (
-                SELECT DISTINCT to_zoid
-                FROM object_ref
-                    JOIN temp_pack_visit USING (zoid)
-                WHERE object_ref.tid >= temp_pack_visit.keep_tid
-            )
+        CREATE UNIQUE INDEX temp_pack_visit_zoid ON temp_pack_visit (zoid);
+        CREATE INDEX temp_pack_keep_tid ON temp_pack_visit (keep_tid)
         """
 
     _script_create_temp_undo = """
@@ -171,6 +182,13 @@ class HistoryPreservingPackUndo(PackUndo):
 
     _script_reset_temp_undo = "DROP TABLE temp_undo"
 
+    _script_find_pack_tid = """
+        SELECT keep_tid
+        FROM pack_object
+        ORDER BY keep_tid DESC
+        LIMIT 1
+        """
+
     _script_transaction_has_data = """
         SELECT tid
         FROM object_state
@@ -213,6 +231,18 @@ class HistoryPreservingPackUndo(PackUndo):
             )
         """
 
+    # See http://www.postgres.cz/index.php/PostgreSQL_SQL_Tricks#Fast_first_n_rows_removing
+    # for = any(array(...)) rationale.
+    _script_delete_empty_transactions_batch = """
+        DELETE FROM transaction
+        WHERE tid = any(array(
+            SELECT tid FROM transaction
+            WHERE packed = %(TRUE)s
+              AND empty = %(TRUE)s
+            LIMIT 1000
+        ))
+        """
+
     def verify_undoable(self, cursor, undo_tid):
         """Raise UndoError if it is not safe to undo the specified txn."""
         stmt = """
@@ -301,14 +331,22 @@ class HistoryPreservingPackUndo(PackUndo):
             AND tid = %(self_tid)s;
 
         -- Copy old states forward.
-        INSERT INTO object_state (zoid, tid, prev_tid, md5, state)
+        INSERT INTO object_state (zoid, tid, prev_tid, md5, state_size, state)
         SELECT temp_undo.zoid, %(self_tid)s, current_object.tid,
-            prev.md5, prev.state
+            md5, COALESCE(state_size, 0), state
         FROM temp_undo
             JOIN current_object ON (temp_undo.zoid = current_object.zoid)
-            LEFT JOIN object_state prev
-                ON (prev.zoid = temp_undo.zoid
-                    AND prev.tid = temp_undo.prev_tid);
+            LEFT JOIN object_state
+                ON (object_state.zoid = temp_undo.zoid
+                    AND object_state.tid = temp_undo.prev_tid);
+
+        -- Copy old blob chunks forward.
+        INSERT INTO blob_chunk (zoid, tid, chunk_num, chunk)
+        SELECT temp_undo.zoid, %(self_tid)s, chunk_num, chunk
+        FROM temp_undo
+            JOIN blob_chunk
+                ON (blob_chunk.zoid = temp_undo.zoid
+                    AND blob_chunk.tid = temp_undo.prev_tid);
 
         -- List the copied states.
         SELECT zoid, prev_tid FROM temp_undo
@@ -338,19 +376,25 @@ class HistoryPreservingPackUndo(PackUndo):
         """
         self.runner.run_script_stmt(cursor, stmt)
         tids = [tid for (tid,) in cursor]
+        log_at = time.time() + 60
         if tids:
             self.on_filling_object_refs()
-            added = 0
-            log.info("discovering references from objects in %d "
-                "transaction(s)" % len(tids))
+            tid_count = len(tids)
+            txns_done = 0
+            log.info("analyzing references from objects in %d new "
+                "transaction(s)", tid_count)
             for tid in tids:
-                added += self._add_refs_for_tid(cursor, tid, get_references)
-                if added >= 10000:
+                self._add_refs_for_tid(cursor, tid, get_references)
+                txns_done += 1
+                now = time.time()
+                if now >= log_at:
                     # save the work done so far
                     conn.commit()
-                    added = 0
-            if added:
-                conn.commit()
+                    log_at = now + 60
+                    log.info(
+                        "transactions analyzed: %d/%d", txns_done, tid_count)
+            conn.commit()
+            log.info("transactions analyzed: %d/%d", txns_done, tid_count)
 
     def _add_refs_for_tid(self, cursor, tid, get_references):
         """Fill object_refs with all states for a transaction.
@@ -359,25 +403,34 @@ class HistoryPreservingPackUndo(PackUndo):
         """
         log.debug("pre_pack: transaction %d: computing references ", tid)
         from_count = 0
+        use_base64 = (self.database_name == 'postgresql')
 
-        stmt = """
-        SELECT zoid, state
-        FROM object_state
-        WHERE tid = %(tid)s
-        """
+        if use_base64:
+            stmt = """
+            SELECT zoid, encode(state, 'base64')
+            FROM object_state
+            WHERE tid = %(tid)s
+            """
+        else:
+            stmt = """
+            SELECT zoid, state
+            FROM object_state
+            WHERE tid = %(tid)s
+            """
         self.runner.run_script_stmt(cursor, stmt, {'tid': tid})
 
-        replace_rows = []
         add_rows = []  # [(from_oid, tid, to_oid)]
         for from_oid, state in cursor:
-            replace_rows.append((from_oid,))
             if hasattr(state, 'read'):
                 # Oracle
                 state = state.read()
             if state:
+                state = str(state)
+                if use_base64:
+                    state = decodestring(state)
                 from_count += 1
                 try:
-                    to_oids = get_references(str(state))
+                    to_oids = get_references(state)
                 except:
                     log.error("pre_pack: can't unpickle "
                         "object %d in transaction %d; state length = %d" % (
@@ -386,10 +439,12 @@ class HistoryPreservingPackUndo(PackUndo):
                 for to_oid in to_oids:
                     add_rows.append((from_oid, tid, to_oid))
 
-        if not self.keep_history:
-            stmt = "DELETE FROM object_ref WHERE zoid = %s"
-            self.runner.run_many(cursor, stmt, replace_rows)
+        # A previous pre-pack may have been interrupted.  Delete rows
+        # from the interrupted attempt.
+        stmt = "DELETE FROM object_ref WHERE tid = %(tid)s"
+        self.runner.run_script_stmt(cursor, stmt, {'tid': tid})
 
+        # Add the new references.
         stmt = """
         INSERT INTO object_ref (zoid, tid, to_zoid)
         VALUES (%s, %s, %s)
@@ -408,7 +463,7 @@ class HistoryPreservingPackUndo(PackUndo):
             "from %d object(s)", tid, to_count, from_count)
         return to_count
 
-    def pre_pack(self, pack_tid, get_references, options):
+    def pre_pack(self, pack_tid, get_references):
         """Decide what to pack.
 
         pack_tid specifies the most recent transaction to pack.
@@ -416,8 +471,8 @@ class HistoryPreservingPackUndo(PackUndo):
         get_references is a function that accepts a pickled state and
         returns a set of OIDs that state refers to.
 
-        options is an instance of relstorage.Options.
-        The options.pack_gc flag indicates whether to run garbage collection.
+        The self.options.pack_gc flag indicates whether
+        to run garbage collection.
         If pack_gc is false, at least one revision of every object is kept,
         even if nothing refers to it.  Packing with pack_gc disabled can be
         much faster.
@@ -425,7 +480,7 @@ class HistoryPreservingPackUndo(PackUndo):
         conn, cursor = self.connmanager.open_for_pre_pack()
         try:
             try:
-                if options.pack_gc:
+                if self.options.pack_gc:
                     log.info("pre_pack: start with gc enabled")
                     self._pre_pack_with_gc(
                         conn, cursor, pack_tid, get_references)
@@ -440,7 +495,7 @@ class HistoryPreservingPackUndo(PackUndo):
                 self.runner.run_script_stmt(cursor, stmt)
                 to_remove = 0
 
-                if options.pack_gc:
+                if self.options.pack_gc:
                     # Pack objects with the keep flag set to false.
                     stmt = """
                     INSERT INTO pack_state (tid, zoid)
@@ -543,8 +598,8 @@ class HistoryPreservingPackUndo(PackUndo):
 
         -- Keep objects that have been revised since pack_tid.
         -- Use temp_pack_visit for temporary state; otherwise MySQL 5 chokes.
-        INSERT INTO temp_pack_visit (zoid)
-        SELECT zoid
+        INSERT INTO temp_pack_visit (zoid, keep_tid)
+        SELECT zoid, 0
         FROM current_object
         WHERE tid > %(pack_tid)s;
 
@@ -559,8 +614,8 @@ class HistoryPreservingPackUndo(PackUndo):
         -- Keep objects that are still referenced by object states in
         -- transactions that will not be packed.
         -- Use temp_pack_visit for temporary state; otherwise MySQL 5 chokes.
-        INSERT INTO temp_pack_visit (zoid)
-        SELECT DISTINCT to_zoid
+        INSERT INTO temp_pack_visit (zoid, keep_tid)
+        SELECT DISTINCT to_zoid, 0
         FROM object_ref
         WHERE tid > %(pack_tid)s;
 
@@ -574,11 +629,21 @@ class HistoryPreservingPackUndo(PackUndo):
         """
         self.runner.run_script(cursor, stmt, {'pack_tid': pack_tid})
 
-        # Set the 'keep' flags in pack_object
-        self._visit_all_references(cursor)
+        # Traverse the graph, setting the 'keep' flags in pack_object
+        self._traverse_graph(cursor)
 
 
-    def pack(self, pack_tid, options, sleep=None, packed_func=None):
+    def _find_pack_tid(self):
+        """If pack was not completed, find our pack tid again"""
+
+        conn, cursor = self.connmanager.open_for_pre_pack()
+        stmt = self._script_find_pack_tid
+        self.runner.run_script_stmt(cursor, stmt)
+        res = [tid for (tid,) in cursor]
+        return res and res[0] or 0
+
+
+    def pack(self, pack_tid, sleep=None, packed_func=None):
         """Pack.  Requires the information provided by pre_pack."""
 
         # Read committed mode is sufficient.
@@ -601,38 +666,51 @@ class HistoryPreservingPackUndo(PackUndo):
                 tid_rows = list(cursor)
                 tid_rows.sort()  # oldest first
 
-                log.info("pack: will pack %d transaction(s)", len(tid_rows))
+                total = len(tid_rows)
+                log.info("pack: will pack %d transaction(s)", total)
 
                 stmt = self._script_create_temp_pack_visit
                 if stmt:
                     self.runner.run_script(cursor, stmt)
 
                 # Hold the commit lock while packing to prevent deadlocks.
-                # Pack in small batches of transactions in order to minimize
-                # the interruption of concurrent write operations.
+                # Pack in small batches of transactions only after we are able
+                # to obtain a commit lock in order to minimize the
+                # interruption of concurrent write operations.
                 start = time.time()
                 packed_list = []
-                self.locker.hold_commit_lock(cursor)
+                counter, lastreport, statecounter = 0, 0, 0
+                # We'll report on progress in at most .1% step increments
+                reportstep = max(total / 1000, 1)
+
+                self._pause_pack_until_lock(cursor, sleep)
                 for tid, packed, has_removable in tid_rows:
                     self._pack_transaction(
                         cursor, pack_tid, tid, packed, has_removable,
                         packed_list)
-                    if time.time() >= start + options.pack_batch_timeout:
+                    counter += 1
+                    if time.time() >= start + self.options.pack_batch_timeout:
                         conn.commit()
                         if packed_func is not None:
                             for oid, tid in packed_list:
                                 packed_func(oid, tid)
+                        statecounter += len(packed_list)
+                        if counter >= lastreport + reportstep:
+                            log.info("pack: packed %d (%.1f%%) transaction(s), "
+                                "affecting %d states",
+                                counter, counter/float(total)*100, 
+                                statecounter)
+                            lastreport = counter / reportstep * reportstep
                         del packed_list[:]
                         self.locker.release_commit_lock(cursor)
-                        self._pause_pack(sleep, options, start)
-                        self.locker.hold_commit_lock(cursor)
+                        self._pause_pack_until_lock(cursor, sleep)
                         start = time.time()
                 if packed_func is not None:
                     for oid, tid in packed_list:
                         packed_func(oid, tid)
                 packed_list = None
 
-                self._pack_cleanup(conn, cursor)
+                self._pack_cleanup(conn, cursor, sleep)
 
             except:
                 log.exception("pack: failed")
@@ -702,30 +780,34 @@ class HistoryPreservingPackUndo(PackUndo):
             tid, state, removed_objects, removed_states)
 
 
-    def _pack_cleanup(self, conn, cursor):
+    def _pack_cleanup(self, conn, cursor, sleep=None):
         """Remove unneeded table rows after packing"""
         # commit the work done so far
         conn.commit()
         self.locker.release_commit_lock(cursor)
-        self.locker.hold_commit_lock(cursor)
         log.info("pack: cleaning up")
 
+        # This section does not need to hold the commit lock, as it only
+        # touches pack-specific tables. We already hold a pack lock for that.
         log.debug("pack: removing unused object references")
         stmt = self._script_pack_object_ref
         self.runner.run_script(cursor, stmt)
 
+        # We need a commit lock when touching the transaction table though.
+        # We'll do it in batches of 1000 rows.
         log.debug("pack: removing empty packed transactions")
-        stmt = """
-        DELETE FROM transaction
-        WHERE packed = %(TRUE)s
-            AND empty = %(TRUE)s
-        """
-        self.runner.run_script_stmt(cursor, stmt)
+        while True:
+            self._pause_pack_until_lock(cursor, sleep)
+            stmt = self._script_delete_empty_transactions_batch
+            self.runner.run_script_stmt(cursor, stmt)
+            deleted = cursor.rowcount
+            conn.commit()
+            self.locker.release_commit_lock(cursor)
+            if deleted < 1000:
+                # Last set of deletions complete
+                break
 
         # perform cleanup that does not require the commit lock
-        conn.commit()
-        self.locker.release_commit_lock(cursor)
-
         log.debug("pack: clearing temporary pack state")
         for _table in ('pack_object', 'pack_state', 'pack_state_tid'):
             stmt = '%(TRUNCATE)s ' + _table
@@ -739,31 +821,11 @@ class MySQLHistoryPreservingPackUndo(HistoryPreservingPackUndo):
     #      http://bugs.mysql.com/bug.php?id=28257
     _script_create_temp_pack_visit = """
         CREATE TEMPORARY TABLE temp_pack_visit (
-            zoid BIGINT NOT NULL,
-            keep_tid BIGINT
+            zoid BIGINT UNSIGNED NOT NULL,
+            keep_tid BIGINT UNSIGNED NOT NULL
         );
         CREATE UNIQUE INDEX temp_pack_visit_zoid ON temp_pack_visit (zoid);
-        CREATE TEMPORARY TABLE temp_pack_child (
-            zoid BIGINT NOT NULL
-        );
-        CREATE UNIQUE INDEX temp_pack_child_zoid ON temp_pack_child (zoid);
-        """
-
-    # Note: UPDATE must be the last statement in the script
-    # because it returns a value.
-    _script_pre_pack_follow_child_refs = """
-        %(TRUNCATE)s temp_pack_child;
-
-        INSERT INTO temp_pack_child
-        SELECT DISTINCT to_zoid
-        FROM object_ref
-            JOIN temp_pack_visit USING (zoid)
-        WHERE object_ref.tid >= temp_pack_visit.keep_tid;
-
-        -- MySQL-specific syntax for table join in update
-        UPDATE pack_object, temp_pack_child SET keep = %(TRUE)s
-        WHERE keep = %(FALSE)s
-            AND pack_object.zoid = temp_pack_child.zoid;
+        CREATE INDEX temp_pack_keep_tid ON temp_pack_visit (keep_tid);
         """
 
     # MySQL optimizes deletion far better when using a join syntax.
@@ -793,6 +855,21 @@ class MySQLHistoryPreservingPackUndo(HistoryPreservingPackUndo):
         WHERE transaction.empty = true
         """
 
+    _script_create_temp_undo = """
+        CREATE TEMPORARY TABLE temp_undo (
+            zoid BIGINT UNSIGNED NOT NULL,
+            prev_tid BIGINT UNSIGNED NOT NULL
+        );
+        CREATE UNIQUE INDEX temp_undo_zoid ON temp_undo (zoid)
+        """
+
+    _script_delete_empty_transactions_batch = """
+        DELETE FROM transaction
+        WHERE packed = %(TRUE)s
+          AND empty = %(TRUE)s
+        LIMIT 1000
+        """
+
 
 class OracleHistoryPreservingPackUndo(HistoryPreservingPackUndo):
 
@@ -808,12 +885,24 @@ class OracleHistoryPreservingPackUndo(HistoryPreservingPackUndo):
     _script_create_temp_undo = None
     _script_reset_temp_undo = "DELETE FROM temp_undo"
 
+    _script_find_pack_tid = """
+        SELECT MAX(keep_tid)
+        FROM pack_object
+        """
+
     _script_transaction_has_data = """
         SELECT DISTINCT tid
         FROM object_state
         WHERE tid = %(tid)s
         """
 
+    _script_delete_empty_transactions_batch = """
+        DELETE FROM transaction
+        WHERE packed = %(TRUE)s
+          AND empty = %(TRUE)s
+          AND rownum <= 1000
+        """
+
 
 class HistoryFreePackUndo(PackUndo):
     implements(IPackUndo)
@@ -832,19 +921,10 @@ class HistoryFreePackUndo(PackUndo):
     _script_create_temp_pack_visit = """
         CREATE TEMPORARY TABLE temp_pack_visit (
             zoid BIGINT NOT NULL,
-            keep_tid BIGINT
+            keep_tid BIGINT NOT NULL
         );
-        CREATE UNIQUE INDEX temp_pack_visit_zoid ON temp_pack_visit (zoid)
-        """
-
-    _script_pre_pack_follow_child_refs = """
-        UPDATE pack_object SET keep = %(TRUE)s
-        WHERE keep = %(FALSE)s
-            AND zoid IN (
-                SELECT DISTINCT to_zoid
-                FROM object_ref
-                    JOIN temp_pack_visit USING (zoid)
-            )
+        CREATE UNIQUE INDEX temp_pack_visit_zoid ON temp_pack_visit (zoid);
+        CREATE INDEX temp_pack_keep_tid ON temp_pack_visit (keep_tid)
         """
 
     def verify_undoable(self, cursor, undo_tid):
@@ -892,22 +972,28 @@ class HistoryFreePackUndo(PackUndo):
             """
             self.runner.run_script_stmt(cursor, stmt)
             oids = [oid for (oid,) in cursor]
+            log_at = time.time() + 60
             if oids:
                 if attempt == 1:
                     self.on_filling_object_refs()
-                added = 0
-                log.info("discovering references from %d objects", len(oids))
+                oid_count = len(oids)
+                oids_done = 0
+                log.info("analyzing references from %d object(s)", oid_count)
                 while oids:
                     batch = oids[:100]
                     oids = oids[100:]
-                    added += self._add_refs_for_oids(
-                        cursor, batch, get_references)
-                    if added >= 10000:
+                    self._add_refs_for_oids(cursor, batch, get_references)
+                    oids_done += len(batch)
+                    now = time.time()
+                    if now >= log_at:
                         # Save the work done so far.
                         conn.commit()
-                        added = 0
-                if added:
-                    conn.commit()
+                        log_at = now + 60
+                        log.info(
+                            "objects analyzed: %d/%d", oids_done, oid_count)
+                conn.commit()
+                log.info(
+                    "objects analyzed: %d/%d", oids_done, oid_count)
             else:
                 # No changes since last pass.
                 break
@@ -917,14 +1003,21 @@ class HistoryFreePackUndo(PackUndo):
 
         Returns the number of references added.
         """
-        to_count = 0
         oid_list = ','.join(str(oid) for oid in oids)
+        use_base64 = (self.database_name == 'postgresql')
 
-        stmt = """
-        SELECT zoid, tid, state
-        FROM object_state
-        WHERE zoid IN (%s)
-        """ % oid_list
+        if use_base64:
+            stmt = """
+            SELECT zoid, tid, encode(state, 'base64')
+            FROM object_state
+            WHERE zoid IN (%s)
+            """ % oid_list
+        else:
+            stmt = """
+            SELECT zoid, tid, state
+            FROM object_state
+            WHERE zoid IN (%s)
+            """ % oid_list
         self.runner.run_script_stmt(cursor, stmt)
 
         add_objects = []
@@ -935,8 +1028,11 @@ class HistoryFreePackUndo(PackUndo):
                 state = state.read()
             add_objects.append((from_oid, tid))
             if state:
+                state = str(state)
+                if use_base64:
+                    state = decodestring(state)
                 try:
-                    to_oids = get_references(str(state))
+                    to_oids = get_references(state)
                 except:
                     log.error("pre_pack: can't unpickle "
                         "object %d in transaction %d; state length = %d" % (
@@ -965,7 +1061,7 @@ class HistoryFreePackUndo(PackUndo):
 
         return len(add_refs)
 
-    def pre_pack(self, pack_tid, get_references, options):
+    def pre_pack(self, pack_tid, get_references):
         """Decide what the garbage collector should delete.
 
         Objects created or modified after pack_tid will not be
@@ -974,11 +1070,10 @@ class HistoryFreePackUndo(PackUndo):
         get_references is a function that accepts a pickled state and
         returns a set of OIDs that state refers to.
 
-        options is an instance of relstorage.Options.
-        The options.pack_gc flag indicates whether to run garbage collection.
-        If pack_gc is false, this method does nothing.
+        The self.options.pack_gc flag indicates whether to run garbage
+        collection.  If pack_gc is false, this method does nothing.
         """
-        if not options.pack_gc:
+        if not self.options.pack_gc:
             log.warning("pre_pack: garbage collection is disabled on a "
                 "history-free storage, so doing nothing")
             return
@@ -1026,11 +1121,19 @@ class HistoryFreePackUndo(PackUndo):
         """
         self.runner.run_script(cursor, stmt, {'pack_tid': pack_tid})
 
-        # Set the 'keep' flags in pack_object
-        self._visit_all_references(cursor)
+        # Traverse the graph, setting the 'keep' flags in pack_object
+        self._traverse_graph(cursor)
 
 
-    def pack(self, pack_tid, options, sleep=None, packed_func=None):
+    def _find_pack_tid(self):
+        """If pack was not completed, find our pack tid again"""
+
+        # pack (below) ignores it's pack_tid argument, so we can safely
+        # return None here
+        return None
+
+
+    def pack(self, pack_tid, sleep=None, packed_func=None):
         """Run garbage collection.
 
         Requires the information provided by pre_pack.
@@ -1047,15 +1150,19 @@ class HistoryFreePackUndo(PackUndo):
                 self.runner.run_script_stmt(cursor, stmt)
                 to_remove = list(cursor)
 
-                log.info("pack: will remove %d object(s)", len(to_remove))
+                total = len(to_remove)
+                log.info("pack: will remove %d object(s)", total)
 
                 # Hold the commit lock while packing to prevent deadlocks.
-                # Pack in small batches of transactions in order to minimize
-                # the interruption of concurrent write operations.
+                # Pack in small batches of transactions only after we are able
+                # to obtain a commit lock in order to minimize the
+                # interruption of concurrent write operations.
                 start = time.time()
                 packed_list = []
-                self.locker.hold_commit_lock(cursor)
+                # We'll report on progress in at most .1% step increments
+                lastreport, reportstep = 0, max(total / 1000, 1)
 
+                self._pause_pack_until_lock(cursor, sleep)
                 while to_remove:
                     items = to_remove[:100]
                     del to_remove[:100]
@@ -1066,15 +1173,19 @@ class HistoryFreePackUndo(PackUndo):
                     self.runner.run_many(cursor, stmt, items)
                     packed_list.extend(items)
 
-                    if time.time() >= start + options.pack_batch_timeout:
+                    if time.time() >= start + self.options.pack_batch_timeout:
                         conn.commit()
                         if packed_func is not None:
                             for oid, tid in packed_list:
                                 packed_func(oid, tid)
                         del packed_list[:]
+                        counter = total - len(to_remove)
+                        if counter >= lastreport + reportstep:
+                            log.info("pack: removed %d (%.1f%%) state(s)",
+                                counter, counter/float(total)*100)
+                            lastreport = counter / reportstep * reportstep
                         self.locker.release_commit_lock(cursor)
-                        self._pause_pack(sleep, options, start)
-                        self.locker.hold_commit_lock(cursor)
+                        self._pause_pack_until_lock(cursor, sleep)
                         start = time.time()
 
                 if packed_func is not None:
@@ -1101,9 +1212,10 @@ class HistoryFreePackUndo(PackUndo):
         # commit the work done so far
         conn.commit()
         self.locker.release_commit_lock(cursor)
-        self.locker.hold_commit_lock(cursor)
         log.info("pack: cleaning up")
 
+        # This section does not need to hold the commit lock, as it only
+        # touches pack-specific tables. We already hold a pack lock for that.
         stmt = """
         DELETE FROM object_refs_added
         WHERE zoid IN (
@@ -1128,30 +1240,10 @@ class MySQLHistoryFreePackUndo(HistoryFreePackUndo):
 
     _script_create_temp_pack_visit = """
         CREATE TEMPORARY TABLE temp_pack_visit (
-            zoid BIGINT NOT NULL,
-            keep_tid BIGINT
+            zoid BIGINT UNSIGNED NOT NULL,
+            keep_tid BIGINT UNSIGNED NOT NULL
         );
         CREATE UNIQUE INDEX temp_pack_visit_zoid ON temp_pack_visit (zoid);
-        CREATE TEMPORARY TABLE temp_pack_child (
-            zoid BIGINT NOT NULL
-        );
-        CREATE UNIQUE INDEX temp_pack_child_zoid ON temp_pack_child (zoid);
-        """
-
-    # Note: UPDATE must be the last statement in the script
-    # because it returns a value.
-    _script_pre_pack_follow_child_refs = """
-        %(TRUNCATE)s temp_pack_child;
-
-        INSERT INTO temp_pack_child
-        SELECT DISTINCT to_zoid
-        FROM object_ref
-            JOIN temp_pack_visit USING (zoid);
-
-        -- MySQL-specific syntax for table join in update
-        UPDATE pack_object, temp_pack_child SET keep = %(TRUE)s
-        WHERE keep = %(FALSE)s
-            AND pack_object.zoid = temp_pack_child.zoid;
         """
 
 
