@@ -13,13 +13,18 @@
 ##############################################################################
 """Database schema installers
 """
+import logging
 from relstorage.adapters.interfaces import ISchemaInstaller
 from ZODB.POSException import StorageError
 from zope.interface import implements
 import re
-import time
 
-relstorage_op_version = '1.4'
+# Versions of the installed stored procedures. Change these when
+# the corresponding code changes.
+oracle_package_version = '1.5A'
+postgresql_proc_version = '1.5B'
+
+log = logging.getLogger("relstorage")
 
 history_preserving_schema = """
 
@@ -80,55 +85,101 @@ history_preserving_schema = """
     oracle:
         CREATE SEQUENCE zoid_seq;
 
-# object_state: All object states in all transactions. Note that md5
-# and state can be null to represent object uncreation.
+# object_state and blob_chunk: All object and blob states in all
+# transactions. Note that md5 and state can be null to represent object
+# uncreation.
 
     postgresql:
         CREATE TABLE object_state (
             zoid        BIGINT NOT NULL,
             tid         BIGINT NOT NULL REFERENCES transaction
                         CHECK (tid > 0),
-            PRIMARY KEY (zoid, tid),
+                        PRIMARY KEY (zoid, tid),
             prev_tid    BIGINT NOT NULL REFERENCES transaction,
             md5         CHAR(32),
+            state_size  BIGINT NOT NULL CHECK (state_size >= 0),
             state       BYTEA
         );
         CREATE INDEX object_state_tid ON object_state (tid);
         CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
 
+        CREATE TABLE blob_chunk (
+            zoid        BIGINT NOT NULL,
+            tid         BIGINT NOT NULL,
+            chunk_num   BIGINT NOT NULL,
+                        PRIMARY KEY (zoid, tid, chunk_num),
+            chunk       OID NOT NULL
+        );
+        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid, tid);
+        CREATE INDEX blob_chunk_loid ON blob_chunk (chunk);
+        ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
+            FOREIGN KEY (zoid, tid)
+            REFERENCES object_state (zoid, tid)
+            ON DELETE CASCADE;
+
     mysql:
         CREATE TABLE object_state (
             zoid        BIGINT NOT NULL,
             tid         BIGINT NOT NULL REFERENCES transaction,
-            PRIMARY KEY (zoid, tid),
+                        PRIMARY KEY (zoid, tid),
+                        CHECK (tid > 0),
             prev_tid    BIGINT NOT NULL REFERENCES transaction,
             md5         CHAR(32) CHARACTER SET ascii,
-            state       LONGBLOB,
-            CHECK (tid > 0)
+            state_size  BIGINT NOT NULL,
+            state       LONGBLOB
         ) ENGINE = InnoDB;
         CREATE INDEX object_state_tid ON object_state (tid);
         CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
 
+        CREATE TABLE blob_chunk (
+            zoid        BIGINT NOT NULL,
+            tid         BIGINT NOT NULL,
+            chunk_num   BIGINT NOT NULL,
+                        PRIMARY KEY (zoid, tid, chunk_num),
+            chunk       LONGBLOB NOT NULL
+        ) ENGINE = InnoDB;
+        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid, tid);
+        ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
+            FOREIGN KEY (zoid, tid)
+            REFERENCES object_state (zoid, tid)
+            ON DELETE CASCADE;
+
     oracle:
         CREATE TABLE object_state (
             zoid        NUMBER(20) NOT NULL,
-            tid         NUMBER(20) NOT NULL REFERENCES transaction
-                        CHECK (tid > 0),
-            PRIMARY KEY (zoid, tid),
+            tid         NUMBER(20) NOT NULL REFERENCES transaction,
+                        PRIMARY KEY (zoid, tid),
+                        CONSTRAINT tid_min CHECK (tid > 0),
             prev_tid    NUMBER(20) NOT NULL REFERENCES transaction,
             md5         CHAR(32),
+            state_size  NUMBER(20) NOT NULL,
+                        CONSTRAINT state_size_min CHECK (state_size >= 0),
             state       BLOB
         );
         CREATE INDEX object_state_tid ON object_state (tid);
         CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
 
+        CREATE TABLE blob_chunk (
+            zoid        NUMBER(20) NOT NULL,
+            tid         NUMBER(20) NOT NULL,
+            chunk_num   NUMBER(20) NOT NULL,
+                        PRIMARY KEY (zoid, tid, chunk_num),
+            chunk       BLOB
+        );
+        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid, tid);
+        ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
+            FOREIGN KEY (zoid, tid)
+            REFERENCES object_state (zoid, tid)
+            ON DELETE CASCADE;
+
 # current_object: Pointers to the current object state
 
     postgresql:
         CREATE TABLE current_object (
             zoid        BIGINT NOT NULL PRIMARY KEY,
             tid         BIGINT NOT NULL,
-            FOREIGN KEY (zoid, tid) REFERENCES object_state
+                        FOREIGN KEY (zoid, tid)
+                            REFERENCES object_state (zoid, tid)
         );
         CREATE INDEX current_object_tid ON current_object (tid);
 
@@ -136,7 +187,8 @@ history_preserving_schema = """
         CREATE TABLE current_object (
             zoid        BIGINT NOT NULL PRIMARY KEY,
             tid         BIGINT NOT NULL,
-            FOREIGN KEY (zoid, tid) REFERENCES object_state (zoid, tid)
+                        FOREIGN KEY (zoid, tid)
+                            REFERENCES object_state (zoid, tid)
         ) ENGINE = InnoDB;
         CREATE INDEX current_object_tid ON current_object (tid);
 
@@ -144,7 +196,8 @@ history_preserving_schema = """
         CREATE TABLE current_object (
             zoid        NUMBER(20) NOT NULL PRIMARY KEY,
             tid         NUMBER(20) NOT NULL,
-            FOREIGN KEY (zoid, tid) REFERENCES object_state
+                        FOREIGN KEY (zoid, tid)
+                            REFERENCES object_state (zoid, tid)
         );
         CREATE INDEX current_object_tid ON current_object (tid);
 
@@ -288,14 +341,21 @@ history_preserving_schema = """
             zoid        NUMBER(20) NOT NULL PRIMARY KEY,
             prev_tid    NUMBER(20) NOT NULL,
             md5         CHAR(32),
-            state       BLOB
+            state       BLOB,
+            blobdata    BLOB
+        ) ON COMMIT DELETE ROWS;
+
+        CREATE GLOBAL TEMPORARY TABLE temp_blob_chunk (
+            zoid        NUMBER(20) NOT NULL,
+            chunk_num   NUMBER(20) NOT NULL,
+            chunk       BLOB
         ) ON COMMIT DELETE ROWS;
 
         # Temporary state during packing: a list of objects
         # whose references need to be examined.
         CREATE GLOBAL TEMPORARY TABLE temp_pack_visit (
             zoid        NUMBER(20) NOT NULL PRIMARY KEY,
-            keep_tid    NUMBER(20)
+            keep_tid    NUMBER(20) NOT NULL
         );
 
         # Temporary state during undo: a list of objects
@@ -339,7 +399,41 @@ history_preserving_init = """
         CREATE SEQUENCE zoid_seq;
 """
 
-oracle_history_preserving_plsql = """
+postgresql_procedures = """
+CREATE OR REPLACE FUNCTION blob_chunk_delete_trigger() RETURNS TRIGGER 
+AS $blob_chunk_delete_trigger$
+    -- Version: %(postgresql_proc_version)s
+    -- Unlink large object data file after blob_chunk row deletion
+    DECLARE
+        cnt integer;
+    BEGIN
+        SELECT count(*) into cnt FROM blob_chunk WHERE chunk=OLD.chunk;
+        IF (cnt = 1) THEN
+            -- Last reference to this oid, unlink
+            PERFORM lo_unlink(OLD.chunk);
+        END IF;
+        RETURN OLD;
+    END;
+$blob_chunk_delete_trigger$ LANGUAGE plpgsql;
+
+CREATE OR REPLACE FUNCTION temp_blob_chunk_delete_trigger() RETURNS TRIGGER
+AS $temp_blob_chunk_delete_trigger$
+    -- Version: %(postgresql_proc_version)s
+    -- Unlink large object data file after temp_blob_chunk row deletion
+    DECLARE
+        cnt integer;
+    BEGIN
+        SELECT count(*) into cnt FROM blob_chunk WHERE chunk=OLD.chunk;
+        IF (cnt = 0) THEN
+            -- No more references to this oid, unlink
+            PERFORM lo_unlink(OLD.chunk);
+        END IF;
+        RETURN OLD;
+    END;
+$temp_blob_chunk_delete_trigger$ LANGUAGE plpgsql;
+""" % globals()
+
+oracle_history_preserving_package = """
 CREATE OR REPLACE PACKAGE relstorage_op AS
     TYPE numlist IS TABLE OF NUMBER(20) INDEX BY BINARY_INTEGER;
     TYPE md5list IS TABLE OF VARCHAR2(32) INDEX BY BINARY_INTEGER;
@@ -358,7 +452,7 @@ END relstorage_op;
 /
 
 CREATE OR REPLACE PACKAGE BODY relstorage_op AS
-/* Version: %s */
+/* Version: %(oracle_package_version)s */
     PROCEDURE store_temp(
         zoids IN numlist,
         prev_tids IN numlist,
@@ -383,16 +477,17 @@ CREATE OR REPLACE PACKAGE BODY relstorage_op AS
             WHERE zoid = zoids(indx)
                 AND tid = tids(indx);
         FORALL indx IN zoids.first..zoids.last
-            INSERT INTO object_state (zoid, tid, prev_tid, md5, state) VALUES
-            (zoids(indx), tids(indx),
-            COALESCE((SELECT tid
-                      FROM current_object
-                      WHERE zoid = zoids(indx)), 0),
-            md5s(indx), states(indx));
+            INSERT INTO object_state
+                (zoid, tid, prev_tid, md5, state_size, state)
+                VALUES (zoids(indx), tids(indx),
+                    COALESCE((SELECT tid
+                        FROM current_object
+                        WHERE zoid = zoids(indx)), 0),
+                md5s(indx), COALESCE(LENGTH(states(indx)), 0), states(indx));
     END restore;
 END relstorage_op;
 /
-""" % relstorage_op_version
+""" % globals()
 
 
 history_free_schema = """
@@ -422,33 +517,78 @@ history_free_schema = """
     oracle:
         CREATE SEQUENCE zoid_seq;
 
-# object_state: All object states in all transactions.
+# object_state and blob_chunk: All object states in all transactions.
 
     postgresql:
         CREATE TABLE object_state (
             zoid        BIGINT NOT NULL PRIMARY KEY,
             tid         BIGINT NOT NULL CHECK (tid > 0),
+            state_size  BIGINT NOT NULL CHECK (state_size >= 0),
             state       BYTEA
         );
         CREATE INDEX object_state_tid ON object_state (tid);
 
+        CREATE TABLE blob_chunk (
+            zoid        BIGINT NOT NULL,
+            chunk_num   BIGINT NOT NULL,
+                        PRIMARY KEY (zoid, chunk_num),
+            tid         BIGINT NOT NULL,
+            chunk       OID NOT NULL
+        );
+        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid);
+        CREATE INDEX blob_chunk_loid ON blob_chunk (chunk);
+        ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
+            FOREIGN KEY (zoid)
+            REFERENCES object_state (zoid)
+            ON DELETE CASCADE;
+
     mysql:
         CREATE TABLE object_state (
             zoid        BIGINT NOT NULL PRIMARY KEY,
             tid         BIGINT NOT NULL,
-            state       LONGBLOB,
-            CHECK (tid > 0)
+                        CHECK (tid > 0),
+            state_size  BIGINT NOT NULL,
+            state       LONGBLOB
         ) ENGINE = InnoDB;
         CREATE INDEX object_state_tid ON object_state (tid);
 
+        CREATE TABLE blob_chunk (
+            zoid        BIGINT NOT NULL,
+            chunk_num   BIGINT NOT NULL,
+                        PRIMARY KEY (zoid, chunk_num),
+            tid         BIGINT NOT NULL,
+            chunk       LONGBLOB NOT NULL
+        ) ENGINE = InnoDB;
+        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid);
+        ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
+            FOREIGN KEY (zoid)
+            REFERENCES object_state (zoid)
+            ON DELETE CASCADE;
+
     oracle:
         CREATE TABLE object_state (
             zoid        NUMBER(20) NOT NULL PRIMARY KEY,
             tid         NUMBER(20) NOT NULL,
+                        CONSTRAINT tid_min CHECK (tid > 0),
+            state_size  NUMBER(20) NOT NULL,
+                        CONSTRAINT state_size_min CHECK (state_size >= 0),
             state       BLOB
         );
         CREATE INDEX object_state_tid ON object_state (tid);
 
+        CREATE TABLE blob_chunk (
+            zoid        NUMBER(20) NOT NULL,
+            chunk_num   NUMBER(20) NOT NULL,
+                        PRIMARY KEY (zoid, chunk_num),
+            tid         NUMBER(20) NOT NULL,
+            chunk       BLOB
+        );
+        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid);
+        ALTER TABLE blob_chunk ADD CONSTRAINT blob_chunk_fk
+            FOREIGN KEY (zoid)
+            REFERENCES object_state (zoid)
+            ON DELETE CASCADE;
+
 # object_ref: A list of referenced OIDs from each object_state. This
 # table is populated as needed during packing.
 
@@ -546,11 +686,17 @@ history_free_schema = """
             state       BLOB
         ) ON COMMIT DELETE ROWS;
 
+        CREATE GLOBAL TEMPORARY TABLE temp_blob_chunk (
+            zoid        NUMBER(20) NOT NULL,
+            chunk_num   NUMBER(20) NOT NULL,
+            chunk       BLOB
+        ) ON COMMIT DELETE ROWS;
+
         # Temporary state during packing: a list of objects
         # whose references need to be examined.
         CREATE GLOBAL TEMPORARY TABLE temp_pack_visit (
             zoid        NUMBER(20) NOT NULL PRIMARY KEY,
-            keep_tid    NUMBER(20)
+            keep_tid    NUMBER(20) NOT NULL
         );
 """
 
@@ -568,7 +714,7 @@ history_free_init = """
         CREATE SEQUENCE zoid_seq;
 """
 
-oracle_history_free_plsql = """
+oracle_history_free_package = """
 CREATE OR REPLACE PACKAGE relstorage_op AS
     TYPE numlist IS TABLE OF NUMBER(20) INDEX BY BINARY_INTEGER;
     TYPE md5list IS TABLE OF VARCHAR2(32) INDEX BY BINARY_INTEGER;
@@ -586,7 +732,7 @@ END relstorage_op;
 /
 
 CREATE OR REPLACE PACKAGE BODY relstorage_op AS
-/* Version: %s */
+/* Version: %(oracle_package_version)s */
     PROCEDURE store_temp(
         zoids IN numlist,
         prev_tids IN numlist,
@@ -608,12 +754,15 @@ CREATE OR REPLACE PACKAGE BODY relstorage_op AS
         FORALL indx IN zoids.first..zoids.last
             DELETE FROM object_state WHERE zoid = zoids(indx);
         FORALL indx IN zoids.first..zoids.last
-            INSERT INTO object_state (zoid, tid, state) VALUES
-            (zoids(indx), tids(indx), states(indx));
+            INSERT INTO object_state (zoid, tid, state_size, state) VALUES (
+                zoids(indx),
+                tids(indx),
+                COALESCE(LENGTH(states(indx)), 0),
+                states(indx));
     END restore;
 END relstorage_op;
 /
-""" % relstorage_op_version
+""" % globals()
 
 
 def filter_script(script, database_name):
@@ -631,6 +780,20 @@ def filter_script(script, database_name):
     return '\n'.join(res)
 
 
+def filter_statements(script, expr):
+    res = []
+    match = False
+    for line in script.splitlines():
+        line = line.strip()
+        if not match and expr.search(line) is not None:
+            match = True
+        if match:
+            res.append(line)
+            if line.rstrip().endswith(';'):
+                match = False
+    return '\n'.join(res)
+
+
 class AbstractSchemaInstaller(object):
 
     # Keep this list in the same order as the schema scripts
@@ -640,7 +803,7 @@ class AbstractSchemaInstaller(object):
         'transaction',
         'new_oid',
         'object_state',
-        'blob_chunk',  # Forward compat with RelStorage 1.5
+        'blob_chunk',
         'current_object',
         'object_ref',
         'object_refs_added',
@@ -648,10 +811,10 @@ class AbstractSchemaInstaller(object):
         'pack_state',
         'pack_state_tid',
         'temp_store',
-        'temp_blob_chunk',  # Forward compat with RelStorage 1.5
+        'temp_blob_chunk',
         'temp_pack_visit',
         'temp_undo',
-        )
+    )
 
     database_name = None  # provided by a subclass
 
@@ -687,6 +850,7 @@ class AbstractSchemaInstaller(object):
                 self.create(cursor)
             else:
                 self.check_compatibility(cursor, tables)
+                self.update_schema(cursor, tables)
         self.connmanager.open_and_call(callback)
 
     def check_compatibility(self, cursor, tables):
@@ -696,14 +860,24 @@ class AbstractSchemaInstaller(object):
                     "Schema mismatch: a history-preserving adapter "
                     "can not connect to a history-free database. "
                     "If you need to convert, use the zodbconvert utility."
-                    )
+                )
         else:
             if 'transaction' in tables and 'current_object' in tables:
                 raise StorageError(
                     "Schema mismatch: a history-free adapter "
                     "can not connect to a history-preserving database. "
                     "If you need to convert, use the zodbconvert utility."
-                    )
+                )
+
+    def update_schema(self, cursor, tables):
+        if not 'blob_chunk' in tables:
+            # Add the blob_chunk table (RelStorage 1.5+)
+            script = filter_script(
+                self.schema_script, self.database_name)
+            expr = (r'(CREATE|ALTER)\s+(GLOBAL TEMPORARY\s+)?(TABLE|INDEX)'
+                '\s+(temp_)?blob_chunk')
+            script = filter_statements(script, re.compile(expr, re.I))
+            self.runner.run_script(cursor, script)
 
     def zap_all(self):
         """Clear all data out of the database."""
@@ -711,14 +885,20 @@ class AbstractSchemaInstaller(object):
             existent = set(self.list_tables(cursor))
             todo = list(self.all_tables)
             todo.reverse()
+            log.debug("Checking tables: %r", todo)
             for table in todo:
+                log.debug("Considering table %s", table)
                 if table.startswith('temp_'):
                     continue
                 if table in existent:
+                    log.debug("Deleting from table %s...", table)
                     cursor.execute("DELETE FROM %s" % table)
+            log.debug("Done deleting from tables.")
             script = filter_script(self.init_script, self.database_name)
             if script:
+                log.debug("Running init script.")
                 self.runner.run_script(cursor, script)
+                log.debug("Done running init script.")
         self.connmanager.open_and_call(callback)
 
     def drop_all(self):
@@ -745,6 +925,29 @@ class PostgreSQLSchemaInstaller(AbstractSchemaInstaller):
             connmanager, runner, keep_history)
         self.locker = locker
 
+    def prepare(self):
+        """Create the database schema if it does not already exist."""
+        def callback(conn, cursor):
+            tables = self.list_tables(cursor)
+            if not 'object_state' in tables:
+                self.create(cursor)
+            else:
+                self.check_compatibility(cursor, tables)
+                self.update_schema(cursor, tables)
+
+            if not self.all_procedures_installed(cursor):
+                self.install_procedures(cursor)
+                if not self.all_procedures_installed(cursor):
+                    raise AssertionError(
+                        "Could not get version information after "
+                        "installing the stored procedures.")
+
+            triggers = self.list_triggers(cursor)
+            if not 'blob_chunk_delete' in triggers:
+                self.install_triggers(cursor)
+
+        self.connmanager.open_and_call(callback)
+
     def create(self, cursor):
         """Create the database tables."""
         super(PostgreSQLSchemaInstaller, self).create(cursor)
@@ -759,6 +962,75 @@ class PostgreSQLSchemaInstaller(AbstractSchemaInstaller):
         cursor.execute("SELECT relname FROM pg_class WHERE relkind = 'S'")
         return [name for (name,) in cursor]
 
+    def list_languages(self, cursor):
+        cursor.execute("SELECT lanname FROM pg_catalog.pg_language")
+        return [name for (name,) in cursor]
+
+    def install_languages(self, cursor):
+        if 'plpgsql' not in self.list_languages(cursor):
+            cursor.execute("CREATE LANGUAGE plpgsql")
+
+    def list_procedures(self, cursor):
+        """Returns {procedure name: version}.  version may be None."""
+        stmt = """
+        SELECT proname, prosrc
+        FROM pg_catalog.pg_namespace n
+        JOIN pg_catalog.pg_proc p ON pronamespace = n.oid
+        JOIN pg_catalog.pg_type t ON prorettype = t.oid
+        WHERE nspname = 'public'
+        """
+        cursor.execute(stmt)
+        res = {}
+        for (name, text) in cursor:
+            version = None
+            match = re.search(r'Version:\s*([0-9a-zA-Z.]+)', text)
+            if match is not None:
+                version = match.group(1)
+            res[name.lower()] = version
+        return res
+
+    def all_procedures_installed(self, cursor):
+        """Check whether all required stored procedures are installed.
+
+        Returns True only if all required procedures are installed and
+        up to date.
+        """
+        expect = [
+            'blob_chunk_delete_trigger',
+            'temp_blob_chunk_delete_trigger',
+        ]
+        current_procs = self.list_procedures(cursor)
+        for proc in expect:
+            if current_procs.get(proc) != postgresql_proc_version:
+                return False
+        return True
+
+    def install_procedures(self, cursor):
+        """Install the stored procedures"""
+        self.install_languages(cursor)
+        cursor.execute(postgresql_procedures)
+
+    def list_triggers(self, cursor):
+        cursor.execute("SELECT tgname FROM pg_trigger")
+        return [name for (name,) in cursor]
+
+    def install_triggers(self, cursor):
+        stmt = """
+        CREATE TRIGGER blob_chunk_delete
+            BEFORE DELETE ON blob_chunk
+            FOR EACH ROW
+            EXECUTE PROCEDURE blob_chunk_delete_trigger()
+        """
+        cursor.execute(stmt)
+
+    def drop_all(self):
+        def callback(conn, cursor):
+            if 'blob_chunk' in self.list_tables(cursor):
+                # Trigger deletion of blob OIDs.
+                cursor.execute("DELETE FROM blob_chunk")
+        self.connmanager.open_and_call(callback)
+        super(PostgreSQLSchemaInstaller, self).drop_all()
+
 
 class MySQLSchemaInstaller(AbstractSchemaInstaller):
     implements(ISchemaInstaller)
@@ -799,31 +1071,28 @@ class OracleSchemaInstaller(AbstractSchemaInstaller):
                 self.create(cursor)
             else:
                 self.check_compatibility(cursor, tables)
+                self.update_schema(cursor, tables)
             packages = self.list_packages(cursor)
-            if packages.get('relstorage_op') != relstorage_op_version:
-                self.install_plsql(cursor)
+            package_name = 'relstorage_op'
+            if packages.get(package_name) != oracle_package_version:
+                self.install_package(cursor)
                 packages = self.list_packages(cursor)
-                if packages.get('relstorage_op') != relstorage_op_version:
+                if packages.get(package_name) != oracle_package_version:
                     raise AssertionError(
                         "Could not get version information after "
-                        "installing the relstorage_op package.")
+                        "installing the %s package." % package_name)
         self.connmanager.open_and_call(callback)
 
-    def install_plsql(self, cursor):
-        """Install the unprivileged PL/SQL package"""
+    def install_package(self, cursor):
+        """Install the package containing stored procedures"""
         if self.keep_history:
-            plsql = oracle_history_preserving_plsql
+            code = oracle_history_preserving_package
         else:
-            plsql = oracle_history_free_plsql
+            code = oracle_history_free_package
 
-        lines = []
-        for line in plsql.splitlines():
-            if line.strip() == '/':
-                # end of a statement
-                cursor.execute('\n'.join(lines))
-                lines = []
-            elif line.strip():
-                lines.append(line)
+        for stmt in code.split('\n/\n'):
+            if stmt.strip():
+                cursor.execute(stmt)
 
     def list_tables(self, cursor):
         cursor.execute("SELECT table_name FROM user_tables")
@@ -834,7 +1103,10 @@ class OracleSchemaInstaller(AbstractSchemaInstaller):
         return [name.lower() for (name,) in cursor]
 
     def list_packages(self, cursor):
-        """Returns {package name: version}.  version may be None."""
+        """List installed stored procedure packages.
+
+        Returns {package name: version}.  version may be None.
+        """
         stmt = """
         SELECT object_name
         FROM user_objects
