@@ -24,13 +24,17 @@ import logging
 import os
 import unittest
 
+
+base_dbname = os.environ.get('RELSTORAGETEST_DBNAME', 'relstoragetest')
+
+
 class UsePostgreSQLAdapter:
     def make_adapter(self):
         from relstorage.adapters.postgresql import PostgreSQLAdapter
         if self.keep_history:
-            db = 'relstoragetest'
+            db = base_dbname
         else:
-            db = 'relstoragetest_hf'
+            db = base_dbname + '_hf'
         return PostgreSQLAdapter(
             dsn='dbname=%s user=relstoragetest password=relstoragetest' % db,
             options=Options(keep_history=self.keep_history),
@@ -42,9 +46,9 @@ class ZConfigTests:
     def checkConfigureViaZConfig(self):
         replica_conf = os.path.join(os.path.dirname(__file__), 'replicas.conf')
         if self.keep_history:
-            dbname = 'relstoragetest'
+            dbname = base_dbname
         else:
-            dbname = 'relstoragetest_hf'
+            dbname = base_dbname + '_hf'
         dsn = (
             "dbname='%s' user='relstoragetest' password='relstoragetest'"
             % dbname)
@@ -56,6 +60,7 @@ class ZConfigTests:
             read-only false
             keep-history %s
             replica-conf %s
+            blob-chunk-size 10MB
             <postgresql>
                 dsn %s
             </postgresql>
@@ -94,6 +99,7 @@ class ZConfigTests:
             self.assertEqual(
                 adapter.connmanager.replica_selector.replica_conf,
                 replica_conf)
+            self.assertEqual(storage._options.blob_chunk_size, 10485760)
         finally:
             db.close()
 
@@ -120,10 +126,10 @@ class HFPostgreSQLFromFile(UsePostgreSQLAdapter, HistoryFreeFromFileStorage):
     pass
 
 db_names = {
-    'data': 'relstoragetest',
-    '1': 'relstoragetest',
-    '2': 'relstoragetest2',
-    'dest': 'relstoragetest2',
+    'data': base_dbname,
+    '1': base_dbname,
+    '2': base_dbname + '2',
+    'dest': base_dbname + '2',
     }
 
 def test_suite():
@@ -153,35 +159,52 @@ def test_suite():
         pass
     else:
         from relstorage.tests.blob.testblob import storage_reusable_suite
-        for keep_history in (False, True):
-            def create_storage(name, blob_dir, keep_history=keep_history):
-                from relstorage.storage import RelStorage
-                from relstorage.adapters.postgresql import PostgreSQLAdapter
-                db = db_names[name]
-                if not keep_history:
-                    db += '_hf'
-                dsn = ('dbname=%s user=relstoragetest '
-                        'password=relstoragetest' % db)
-                adapter = PostgreSQLAdapter(
-                    dsn=dsn, options=Options(keep_history=keep_history))
-                storage = RelStorage(adapter, name=name, create=True,
-                    blob_dir=os.path.abspath(blob_dir))
-                storage.zap_all()
-                return storage
+        from relstorage.tests.util import shared_blob_dir_choices
+        for shared_blob_dir in shared_blob_dir_choices:
+            for keep_history in (False, True):
+                def create_storage(name, blob_dir,
+                        shared_blob_dir=shared_blob_dir,
+                        keep_history=keep_history, **kw):
+                    from relstorage.storage import RelStorage
+                    from relstorage.adapters.postgresql import PostgreSQLAdapter
+                    db = db_names[name]
+                    if not keep_history:
+                        db += '_hf'
+                    dsn = ('dbname=%s user=relstoragetest '
+                            'password=relstoragetest' % db)
+                    options = Options(
+                        keep_history=keep_history,
+                        shared_blob_dir=shared_blob_dir,
+                        blob_dir=os.path.abspath(blob_dir),
+                        **kw)
+                    adapter = PostgreSQLAdapter(dsn=dsn, options=options)
+                    storage = RelStorage(adapter, name=name, options=options)
+                    storage.zap_all()
+                    return storage
 
-            if keep_history:
-                prefix = 'HPPostgreSQL'
-                pack_test_name = 'blob_packing.txt'
-            else:
-                prefix = 'HFPostgreSQL'
-                pack_test_name = 'blob_packing_history_free.txt'
+                prefix = 'PostgreSQL%s%s' % (
+                    (shared_blob_dir and 'Shared' or 'Unshared'),
+                    (keep_history and 'WithHistory' or 'NoHistory'),
+                )
 
-            suite.addTest(storage_reusable_suite(
-                prefix, create_storage,
-                test_blob_storage_recovery=True,
-                test_packing=True,
-                test_undo=keep_history,
-                pack_test_name=pack_test_name,
+                # If the blob directory is a cache, don't test packing,
+                # since packing can not remove blobs from all caches.
+                test_packing = shared_blob_dir
+
+                if keep_history:
+                    pack_test_name = 'blob_packing.txt'
+                else:
+                    pack_test_name = 'blob_packing_history_free.txt'
+
+                suite.addTest(storage_reusable_suite(
+                    prefix, create_storage,
+                    test_blob_storage_recovery=True,
+                    test_packing=test_packing,
+                    test_undo=keep_history,
+                    pack_test_name=pack_test_name,
+                    test_blob_cache=(not shared_blob_dir),
+                    # PostgreSQL blob chunks are max 2GB in size
+                    large_blob_size=(not shared_blob_dir) and (1<<31) + 100,
                 ))
 
     return suite
