@@ -40,7 +40,20 @@ def main(argv=sys.argv):
     parser.add_option(
         "-d", "--days", dest="days", default="0",
         help="Days of history to keep (default 0)",
-        )
+    )
+    parser.add_option(
+        "--prepack", dest="prepack", default=False,
+        action="store_true",
+        help="Perform only the pre-pack preparation stage of a pack. "
+        "(Only works with some storage types)",
+    )
+    parser.add_option(
+        "--use-prepack-state", dest="reuse_prepack", default=False,
+        action="store_true",
+        help="Skip the preparation stage and go straight to packing. "
+        "Requires that a pre-pack has been run, or that packing was aborted "
+        "before it was completed.",
+    )
     options, args = parser.parse_args(argv[1:])
 
     if len(args) != 1:
@@ -59,7 +72,13 @@ def main(argv=sys.argv):
         log.info("Opening %s...", name)
         storage = s.open()
         log.info("Packing %s.", name)
-        storage.pack(t, ZODB.serialize.referencesf)
+        if options.prepack or options.reuse_prepack:
+            storage.pack(t, ZODB.serialize.referencesf,
+                prepack_only=options.prepack,
+                skip_prepack=options.reuse_prepack)
+        else:
+            # Be non-relstorage Storages friendly
+            storage.pack(t, ZODB.serialize.referencesf)
         storage.close()
         log.info("Packed %s.", name)
 
